import py
from sympy import *
from sympy.numerics import *
from sympy.numerics.functions import *
from sympy.numerics.quad import *

def test_nintegrate():
    from operator import abs    # workaround abs / sympy.abs conflict
    Float.store()
    Float.setdps(20)
    pi_ = pi_float()
    assert nintegrate(lambda x: sin(x), 0, pi_).ae(2)
    assert nintegrate(lambda x: abs(sin(x)), 0, 10*pi_).ae(20)
    assert nintegrate(lambda x: sin(x), 0, 10*pi_).ae(0)
    assert nintegrate(lambda x: 4/(1+x**2), 0, 1).ae(pi_)
    assert nintegrate(lambda x: 4*sqrt(1-x**2), 0, 1).ae(pi_)
    Float.revert()

def test_nintegrate_infinite():
    Float.store()
    Float.setdps(15)
    pi_ = pi_float()
    assert nintegrate(lambda x: 4/(1+x**2), 1, oo).ae(pi_)
    A = nintegrate(lambda x: 2 * exp(-x**2), 0, oo)
    B = nintegrate(lambda x: 2 * exp(-x**2), -oo, 0)
    C = nintegrate(lambda x: 2 * exp(-x**2), -oo, oo)
    D = nintegrate(lambda x: 2 * exp(-x**2), 1, oo)
    E = nintegrate(lambda x: 2 * exp(-x**2), -1, oo)
    F = nintegrate(lambda x: 2 * exp(-x**2), -oo, -1)
    G = nintegrate(lambda x: 2 * exp(-x**2), -oo, 1)
    assert A.ae(pi_ ** 0.5)
    assert A.ae(B)
    assert C.ae(2*B)
    assert D.ae(0.27880558528066197650)
    assert E.ae(3.2661021165303700781)
    assert F.ae(D)
    assert G.ae(E)
    Float.revert()

def test_tanhsinh():
    Float.store()
    Float.setdps(15)
    assert nintegrate(lambda x: x**3, -3, 2, method=1).ae(-16.25)
    assert nintegrate(lambda x: 2/(1+x**2), -1, 1, method=1).ae(pi_float())
    assert nintegrate(lambda x: 2/(1+x**2), 0, oo, method=1).ae(pi_float())
    assert nintegrate(lambda x: exp(-x), 0, oo, method=1).ae(1)
    assert nintegrate(lambda x: 2*exp(-x**2), 0, oo, method=1).ae(sqrt(pi_float()))
    Float.revert()



"""
Decimal math functions.

Heavily copied from http://code.google.com/p/dmath/, generalized pow to non-integer exponents.

"""

__all__ = ['acos', 'asin', 'atan', 'atan2', 'ceiling', 'cos', 'cosh', 'degrees',
           'e', 'exp', 'floor', 'golden_ratio', 'hypot', 'log', 'log10', 'pi',
           'pow', 'radians', 'sign', 'sin', 'sinh', 'sqrt', 'tan', 'tanh',
           'cot','coth']

import math, decimal

from decimal import getcontext
from decimal import Decimal as D

def pi():
    """Compute Pi to the current precision."""
    getcontext().prec += 2
    lasts, t, s, n, na, d, da = 0, D(3), 3, 1, 0, 0, 24
    while s != lasts:
        lasts = s
        n, na = n + na, na + 8
        d, da = d + da, da + 32
        t = (t * n) / d
        s += t
    getcontext().prec -= 2
    return +s

def e():
    """Compute the base of the natural logarithm to the current precision."""
    return exp(D(1))

def golden_ratio():
    """Calculate the golden ratio to the current precision."""
    return  +((1 + D(5).sqrt()) / 2)


def exp(x):
    """Return e raised to the power of x.
    """
    getcontext().prec += 2
    i, lasts, s, fact, num = 0, 0, 1, 1, 1
    while s != lasts:
        lasts = s
        i += 1
        fact *= i
        num *= x
        s += num / fact
    getcontext().prec -= 2
    return +s

def cos(x):
    """Return the cosine of x as measured in radians.
    """
    getcontext().prec += 2
    i, lasts, s, fact, num, sign = 0, 0, 1, 1, 1, 1
    while s != lasts:
        lasts = s
        i += 2
        fact *= i * (i - 1)
        num *= x * x
        sign *= -1
        s += num / fact * sign
    getcontext().prec -= 2
    return +s

def sin(x):
    """Return the sine of x as measured in radians.
    """
    getcontext().prec += 2
    i, lasts, s, fact, num, sign = 1, 0, x, 1, x, 1
    while s != lasts:
        lasts = s
        i += 2
        fact *= i * (i - 1)
        num *= x * x
        sign *= -1
        s += num / fact * sign
    getcontext().prec -= 2
    return +s

def cosh(x):
    """Return the hyperbolic cosine of Decimal x."""
    if x == 0:
        return D(1)

    getcontext().prec += 2
    i, lasts, s, fact, num = 0, 0, 1, 1, 1
    while s != lasts:
        lasts = s
        i += 2
        num *= x * x
        fact *= i * (i - 1)
        s += num / fact
    getcontext().prec -= 2
    return +s

def sinh(x):
    """Return the hyperbolic sine of Decimal x."""
    if x == 0:
        return D(0)

    getcontext().prec += 2
    i, lasts, s, fact, num = 1, 0, x, 1, x
    while s != lasts:
        lasts = s
        i += 2
        num *= x * x
        fact *= i * (i - 1)
        s += num / fact
    getcontext().prec -= 2
    return +s

def asin(x):
    """Return the arc sine (measured in radians) of Decimal x."""
    if abs(x) > 1:
        raise ValueError("Domain error: asin accepts -1 <= x <= 1")

    if x == -1:
        return pi() / -2
    elif x == 0:
        return D(0)
    elif x == 1:
        return pi() / 2

    return atan2(x, D.sqrt(1 - x ** 2))

def acos(x):
    """Return the arc cosine (measured in radians) of Decimal x."""
    if abs(x) > 1:
        raise ValueError("Domain error: acos accepts -1 <= x <= 1")

    if x == -1:
        return pi()
    elif x == 0:
        return pi() / 2
    elif x == 1:
        return D(0)

    return pi() / 2 - atan2(x, D.sqrt(1 - x ** 2))

def tan(x):
    """Return the tangent of Decimal x (measured in radians)."""
    return +(sin(x) / cos(x))

def tanh(x):
    """Return the hyperbolic tangent of Decimal x."""
    return +(sinh(x) / cosh(x))

def cot(x):
    """Return the cotangent of Decimal x (measured in radians)."""
    return +(cos(x) / sin(x))

def coth(x):
    """Return the hyperbolic cotangent of Decimal x."""
    return +(cosh(x) / sinh(x))

def atan(x):
    """Return the arc tangent (measured in radians) of Decimal x."""
    if x == D('-Inf'):
        return pi() / -2
    elif x == 0:
        return D(0)
    elif x == D('Inf'):
        return pi() / 2

    if x < -1:
        c = pi() / -2
        x = 1 / x
    elif x > 1:
        c = pi() / 2
        x = 1 / x
    else:
        c = 0

    getcontext().prec += 2
    x_squared = x ** 2
    y = x_squared / (1 + x_squared)
    y_over_x = y / x
    i, lasts, s, coeff, num = D(0), 0, y_over_x, 1, y_over_x
    while s != lasts:
        lasts = s
        i += 2
        coeff *= i / (i + 1)
        num *= y
        s += coeff * num
    if c:
        s = c - s
    getcontext().prec -= 2
    return +s

def sign(x):
    """Return -1 for negative numbers and 1 for positive numbers."""
    return 2 * D(x >= 0) - 1

def atan2(y, x):
    """Return the arc tangent (measured in radians) of y/x.
    Unlike atan(y/x), the signs of both x and y are considered.
    """
    abs_y = abs(y)
    abs_x = abs(x)
    y_is_real = abs_y != D('Inf')

    if x:
        if y_is_real:
            a = y and atan(y / x) or D(0)
            if x < 0:
                a += sign(y) * pi()
            return a
        elif abs_y == abs_x:
            x = sign(x)
            y = sign(y)
            return pi() * (D(2) * abs(x) - x) / (D(4) * y)
    if y:
        return atan(sign(y) * D('Inf'))
    elif x < 0:
        return sign(y) * pi()
    else:
        return D(0)

def log(x, base=None):
    """log(x[, base]) -> the logarithm of Decimal x to the given Decimal base.
    If the base not specified, returns the natural logarithm (base e) of x.
    """
    if x < 0:
        return D('NaN')
    elif base == 1:
        raise ValueError("Base was 1!")
    elif x == base:
        return D(1)
    elif x == 0:
        return D('-Inf')

    getcontext().prec += 2

    if x < 1 or x > 10:
        # Consider x = a / 10^b, with |b| > 1 and 1 < a < 10. As long as our
        # log calculation is fast for 1 <= x <= 10, then we can consider x
        # in this form and then log(x) = log(a/10^b) = log(a) - b*log(10)
        r1 = list(x.as_tuple())
        r1[2] = 1 - len(r1[1])
        a = D(tuple(r1))

        r2 = list(x.as_tuple())
        b = 1 - r2[2] - len(r2[1])

        ret = log(a, base) - b*log(D(10), base)
        getcontext().prec -= 2
        return +ret

    if base is None:
        log_base = 1
        approx = math.log(x)
    else:
        log_base = log(base)
        approx = math.log(x, base)

    lasts, s = 0, D(repr(approx))
    while lasts != s:
        lasts = s
        s = s - 1 + x / exp(s)
    s /= log_base

    getcontext().prec -= 2
    return +s

def log10(x):
    """log10(x) -> the base 10 logarithm of Decimal x."""
    return log(x, D(10))

def pow(x, y):
    """pow(x,y) -> x ** y."""
    if isinstance(y,(int,long)):
        return D.__pow__(x, y)
    if isinstance(y,D) and y==D('0.5'):
        return sqrt(x)
    return exp(y * log(x)) # this is slow

sqrt = D.sqrt

def degrees(x):
    """degrees(x) -> converts Decimal angle x from radians to degrees"""
    return +(x * 180 / pi())

def radians(x):
    """radians(x) -> converts Decimal angle x from degrees to radians"""
    return +(x * pi() / 180)

def ceiling(x):
    """Return the smallest integral value >= x."""
    return x.to_integral(rounding=decimal.ROUND_CEILING)

def floor(x):
    """Return the largest integral value <= x."""
    return x.to_integral(rounding=decimal.ROUND_FLOOR)

def hypot(x, y):
    """Return the Euclidean distance, sqrt(x*x + y*y)."""
    return sqrt(x * x + y * y)

# EOF



from sympy import *
from sympy.core import sympify
from sympy.polynomials import Polynomial
from float_ import Float, ComplexFloat
import functions
import constants
from utils_ import bitcount


def polyfunc(expr, derivative=False):
    """
    Convert a SymPy expression representing a univariate polynomial
    into a function for numerical evaluation using Floats /
    ComplexFloats.

        >>> x = Symbol('x')
        >>> polyfunc(x**3 + 4)(2)
        Float('12')

    If derivative=True is set, the evaluation function evaluates both
    the polynomial and its derivative at the given point and returns
    the two values as a tuple.
    """
    poly = Polynomial(expr)
    degree = poly.coeffs[0][1]
    coeffs = [0] * int(degree + 1)
    for c, e in poly.coeffs:
        coeffs[int(e)] = evalf(c)
    def g(x):
        x = evalf(x)
        p = coeffs[int(degree)]
        q = 0
        for i in xrange(degree-1, -1, -1):
            if derivative:
                q = p + x*q
            p = coeffs[i] + x*p
        if derivative:
            return p, q
        else:
            return p
    return g


def evalf(expr):
    """
    evalf(expr) attempts to evaluate a SymPy expression to a Float or
    ComplexFloat with an error smaller than 10**(-Float.getdps())
    """

    if isinstance(expr, (Float, ComplexFloat)):
        return expr
    elif isinstance(expr, (int, float)):
        return Float(expr)
    elif isinstance(expr, complex):
        return ComplexFloat(expr)

    expr = sympify(expr)

    if isinstance(expr, (Rational)):
        y = Float(expr)
    elif isinstance(expr, Real):
        y = Float(str(expr))

    elif expr is I:
        y = ComplexFloat(0,1)

    elif expr is pi:
        y = constants.pi_float()

    elif expr is E:
        y = functions.exp(1)

    elif isinstance(expr, Mul):
        factors = expr[:]
        workprec = Float.getprec() + 1 + len(factors)
        Float.store()
        Float.setprec(workprec)
        y = Float(1)
        for f in factors:
            y *= evalf(f)
        Float.revert()

    elif isinstance(expr, Pow):
        base, expt = expr[:]
        workprec = Float.getprec() + 8 # may need more
        Float.store()
        Float.setprec(workprec)
        base = evalf(base)
        expt = evalf(expt)
        if expt == 0.5:
            y = functions.sqrt(base)
        else:
            y = functions.exp(functions.log(base) * expt)
        Float.revert()

    elif isinstance(expr, C.exp):
        Float.store()
        Float.setprec(Float.getprec() + 3)
        #XXX: how is it possible, that this works:
        x = evalf(expr[0])
        #and this too:
        #x = evalf(expr[1])
        #?? (Try to uncomment it and you'll see)
        y = functions.exp(x)
        Float.revert()

    elif isinstance(expr, Add):
        # TODO: this doesn't yet work as it should.
        # We need some way to handle sums whose results are
        # very close to 0, and when necessary, repeat the
        # summation with higher precision
        reqprec = Float.getprec()
        Float.store()
        Float.setprec(10)
        terms = expr[:]
        approxterms = [abs(evalf(x)) for x in terms]
        min_mag = min(x.exp for x in approxterms)
        max_mag = max(x.exp+bitcount(x.man) for x in approxterms)
        Float.setprec(reqprec - 10 + max_mag - min_mag + 1 + len(terms))
        workprec = Float.getdps()
        y = 0
        for t in terms:
            y += evalf(t)
        Float.revert()

    else:
        # print expr, expr.__class__
        raise NotImplementedError

    # print expr, y

    return +y



"""
Commonly used mathematical constants (pi, log(2), log(10), gamma, ...)

There are two functions for each constant: const_fixed(prec) returns
a big integer n such that const ~= n * 2**-prec, and const_float()
returns Float(const), rounded to the current Float working precision
"""

from math import log as _clog
from float_ import Float
from utils_ import global_options

# Only re-compute a constant if the precision level is raised
def _constmemo(f):
    f.memo_prec = -1
    f.memo_val = None
    def calc(prec):
        if prec == f.memo_prec: return f.memo_val
        if prec < f.memo_prec: return f.memo_val >> (f.memo_prec-prec)
        f.memo_val = f(prec)
        f.memo_prec = prec
        return f.memo_val
    return calc

# Evaluate a Machin-like formula, i.e., a rational combination of
# of acot(n) or acoth(n) for specific integer values of n
def _machin(coefs, prec, hyperbolic=False):
    prec += 10
    def acot(x):
        # Use the basic series expansion for atan/acot, optimized
        # for integer arguments
        s = w = (1<<prec)//x; x2 = x*x;  n = 3
        while 1:
            w //= x2
            term = w // n
            if not term: break
            if hyperbolic or n & 2 == 0: s += term
            else: s -= term
            n += 2
        return s
    s = 0
    for a, b in coefs:
        s += a * acot(b)
    return s >> 10

#----------------------------------------------------------------------
# Pi
#

"""
At low precision, pi can be calculated easily using Machin's formula
pi = 16*acot(5)-4*acot(239).

For high precision, we use the Brent-Salamin algorithm based on the
arithmetic-geometric mean. See for example Wikipedia
(http://en.wikipedia.org/wiki/Brent-Salamin_algorithm)
or "Pi and the AGM" by Jonathan and Peter Borwein (Wiley, 1987).
The algorithm (as stated in the Wikipedia article) consists of
setting

  a_0 = 1;  b_0 = 1/sqrt(2);  t_0 = 1/4;  p_0 = 1

and computing

  a_{n+1} = (a_n + b_n)/2
  b_{n+1} = sqrt(a_n * b_n)
  t_{n+1} = t_n - p_n*(a_n - a_{n+1})**2
  p_{n+1} = 2*p_n

for n = 0, 1, 2, 3, ..., after which the approximation is given by

  pi ~= (a_n + b_n)**2 / (4*t_n).

Each step roughly doubles the number of correct digits.
"""

def _pi_agm(prec):
    from functions import _sqrt_fixed2
    prec += 50
    a = 1 << prec
    if "verbose" in global_options:
        print "  computing initial square root..."
    b = _sqrt_fixed2(a >> 1, prec)
    t = a >> 2
    p = 1
    step = 1
    while 1:
        an = (a+b)>>1
        adiff = a - an
        if "verbose" in global_options:
            base = global_options.get("verbose_base", 10)
            try:
                logdiff = _clog(adiff, base)
            except ValueError:
                logdiff = 0
            digits = int(prec/_clog(base,2) - logdiff)
            print "  iteration", step, ("(accuracy ~= %i base-%i digits)" % (digits, base))
        if p > 16 and abs(adiff) < 1000:
            break
        prod = (a*b)>>prec
        b = _sqrt_fixed2(prod, prec)
        t = t - p*((adiff**2) >> prec)
        p = 2*p
        a = an
        step += 1
    if "verbose" in global_options:
        print "  final division"
    return ((((a+b)**2) >> 2) // t) >> 50

@_constmemo
def pi_fixed(prec):
    if prec < 10000:
        return _machin([(16, 5), (-4, 239)], prec)
    else:
        return _pi_agm(prec)

def pi_float():
    prec = Float._prec + 5
    return Float((pi_fixed(prec), -prec))

#----------------------------------------------------------------------
# Logarithms
#

# Logarithms of integers can be computed easily using
# Machin-like formulas

@_constmemo
def log2_fixed(prec):
    return _machin([(18, 26), (-2, 4801), (8, 8749)], prec, True)

def log2_float():
    prec = Float._prec + 5
    return Float((log2_fixed(prec), -prec))

@_constmemo
def log10_fixed(prec):
    return _machin([(46, 31), (34, 49), (20, 161)], prec, True)

def log10_float():
    prec = Float._prec + 5
    return Float((log10_fixed(prec), -prec))

#----------------------------------------------------------------------
# Other constants
#

"""
Euler's constant (gamma) is computed using the Brent-McMillan formula,
gamma ~= A(n)/B(n) - log(n), where

  A(n) = sum_{k=0,1,2,...} (n**k / k!)**2 * H(k)
  B(n) = sum_{k=0,1,2,...} (n**k / k!)**2
  H(k) = 1 + 1/2 + 1/3 + ... + 1/k

The error is bounded by O(exp(-4n)). Choosing n to be a power
of two, 2**p, the logarithm becomes particularly easy to calculate.

Reference:
Xavier Gourdon & Pascal Sebah, The Euler constant: gamma
http://numbers.computation.free.fr/Constants/Gamma/gamma.pdf
"""

@_constmemo
def gamma_fixed(prec):
    # XXX: may need even more extra precision
    prec += 30
    # choose p such that exp(-4*(2**p)) < 2**-n
    p = int(_clog((prec/4) * _clog(2), 2)) + 1
    n = 1<<p; r=one=1<<prec
    H, A, B, npow, k, d = 0, 0, 0, 1, 1, 1
    while r:
        A += (r * H) >> prec
        B += r
        r = r * (n*n) // (k*k)
        H += one // k
        k += 1
    S = ((A<<prec) // B) - p*log2_fixed(prec)
    return S >> 30

def gamma_float():
    prec = Float._prec + 5
    return Float((gamma_fixed(prec), -prec))
